@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (c) 1998-2005 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (c) 1998-2006 Michal Trojnara <Michal.Trojnara@mirt.net>
  *                 All Rights Reserved
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -26,11 +26,51 @@
 
 /* #define DEBUG_STACK_SIZE */
 
-#ifndef VERSION
-#define VERSION "4.11"
+/**************************************** Platform */
+
+#ifdef USE_WIN32
+#define USE_IPv6
 #endif
 
-#ifndef USE_WIN32
+#ifdef _WIN32_WCE
+#define USE_WIN32
+typedef int socklen_t;
+#define EINTR WSAEINTR
+#define EMFILE WSAEMFILE
+#endif
+
+#ifdef USE_WIN32
+#define HAVE_OPENSSL
+#define HAVE_OSSL_ENGINE_H
+#endif
+
+/**************************************** Common constants */
+
+#ifndef VERSION
+#define VERSION "4.15"
+#endif
+
+/* CPU stack size */
+#define STACK_SIZE 65536
+
+/* I/O buffer size */
+#define BUFFSIZE        16384
+
+/* Length of strings (including the terminating '\0' character) */
+#define STRLEN          256
+
+/* IP address and TCP port textual representation length */
+#define IPLEN           128
+
+/* How many bytes of random input to read from files for PRNG */
+/* OpenSSL likes at least 128 bits, so 64 bytes seems plenty. */
+#define RANDOM_BYTES    64
+
+/**************************************** Generic headers */
+
+#ifdef __vms
+#include <starlet.h>
+#endif /* __vms */
 
 /* For nsr-tandem-nsk architecture */
 #ifdef __TANDEM
@@ -41,7 +81,6 @@
 #ifdef USE_UCONTEXT
 #define __MAKECONTEXT_V2_SOURCE
 #include <ucontext.h>
-#define DEBUG_STACK_SIZE
 #endif
 
 #ifdef USE_PTHREAD
@@ -56,13 +95,11 @@
 #define USE_LIBWRAP
 #endif
 
-#endif /* !USE_WIN32 */
-
 /* Must be included before sys/stat.h for Ultrix */
 #include <sys/types.h>   /* u_short, u_long */
-
 /* General headers */
 #include <stdio.h>
+/* Must be included before sys/stat.h for Ultrix */
 #include <errno.h>
 #include <stdlib.h>
 #include <stdarg.h>      /* va_ */
@@ -70,38 +107,31 @@
 #include <ctype.h>       /* isalnum */
 #include <time.h>
 #include <sys/stat.h>    /* stat */
+#include <setjmp.h>
 
-#if defined (USE_WIN32) || defined (__vms)
-#define LOG_EMERG       0
-#define LOG_ALERT       1
-#define LOG_CRIT        2
-#define LOG_ERR         3
-#define LOG_WARNING     4
-#define LOG_NOTICE      5
-#define LOG_INFO        6
-#define LOG_DEBUG       7
-#endif /* defined (USE_WIN32) || defined (__vms) */
+/**************************************** WIN32 headers */
 
 #ifdef USE_WIN32
 
+#ifndef HOST
 #ifdef __MINGW32__
 #define HOST "x86-pc-mingw32-gnu"
 #else
 #define HOST "x86-pc-unknown"
 #endif
+#endif
 
 typedef unsigned char u8;
 typedef unsigned short u16;
 typedef unsigned long u32;
 
+#define HAVE_SNPRINTF
+#define snprintf _snprintf
 #define HAVE_VSNPRINTF
 #define vsnprintf _vsnprintf
-/*  Already defined for mingw, perhaps others
-int _vsnprintf(char *, int, char *, ...);
-*/
 #define strcasecmp _stricmp
 #define strncasecmp _strnicmp
-#define sleep(c) _sleep(c)
+#define sleep(c) Sleep(1000*(c))
 
 #define exit(c) exit_stunnel(c)
 
@@ -115,9 +145,12 @@ int _vsnprintf(char *, int, char *, ...);
 #define __USE_W32_SOCKETS
 
 /* Winsock2 header for IPv6 definitions */
+#ifdef _WIN32_WCE
+#include <winsock.h>
+#else
 #include <winsock2.h>
 #include <ws2tcpip.h>
-
+#endif
 #include <windows.h>
 
 #define ECONNRESET WSAECONNRESET
@@ -130,10 +163,13 @@ int _vsnprintf(char *, int, char *, ...);
 #define EINVAL WSAEINVAL
 
 #include <process.h>     /* _beginthread */
+#include <tchar.h>
 
 #define NO_IDEA
 #define OPENSSL_NO_IDEA
 
+/**************************************** non-WIN32 headers */
+
 #else /* USE_WIN32 */
 
 #if SIZEOF_UNSIGNED_CHAR == 1
@@ -189,13 +225,19 @@ typedef unsigned long u32;
 #ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>  /* for aix */
 #endif
+
+#ifndef BROKEN_POLL
 #ifdef HAVE_POLL_H
 #include <poll.h>
+#define USE_POLL
 #else /* HAVE_POLL_H */
 #ifdef HAVE_SYS_POLL_H
 #include <sys/poll.h>
+#define USE_POLL
 #endif /* HAVE_SYS_POLL_H */
 #endif /* HAVE_POLL_H */
+#endif /* BROKEN_POLL */
+
 #ifdef HAVE_SYS_FILIO_H
 #include <sys/filio.h>   /* for FIONBIO */
 #endif
@@ -208,7 +250,6 @@ typedef unsigned long u32;
 #endif
 #include <fcntl.h>
 
-    /* Networking headers */
 #include <netinet/in.h>  /* struct sockaddr_in */
 #include <sys/socket.h>  /* getpeername */
 #include <arpa/inet.h>   /* inet_ntoa */
@@ -244,14 +285,15 @@ extern char *sys_errlist[];
 
 #endif /* USE_WIN32 */
 
-/* OpenSSL headers */
+/**************************************** OpenSSL headers */
+
 #ifdef HAVE_OPENSSL
 #include <openssl/lhash.h>
 #include <openssl/ssl.h>
 #include <openssl/err.h>
 #include <openssl/crypto.h> /* for CRYPTO_* and SSLeay_version */
 #include <openssl/rand.h>
-#if (SSLEAY_VERSION_NUMBER >= 0x00907000L) && defined(HAVE_OSSL_ENGINE_H)
+#ifdef HAVE_OSSL_ENGINE_H
 #include <openssl/engine.h>
 #endif
 #else
@@ -261,21 +303,7 @@ extern char *sys_errlist[];
 #include <crypto.h> /* for CRYPTO_* and SSLeay_version */
 #endif
 
-/* CPU stack size */
-#define STACK_SIZE 65536
-
-/* I/O buffer size */
-#define BUFFSIZE        16384
-
-/* Length of strings (including the terminating '\0' character) */
-#define STRLEN          256
-
-/* IP address and TCP port textual representation length */
-#define IPLEN           128
-
-/* How many bytes of random input to read from files for PRNG */
-/* OpenSSL likes at least 128 bits, so 64 bytes seems plenty. */
-#define RANDOM_BYTES    64
+/**************************************** Other defines */
 
 /* Safe copy for strings declarated as char[STRLEN] */
 #define safecopy(dst, src) \
@@ -290,7 +318,7 @@ extern char *sys_errlist[];
     do {char *p; for(p=(s); *p; p++) if(!isalnum((int)*p)) *p='.';} while(0)
 
 /* Some definitions for IPv6 support */
-#if defined(USE_IPv6) || defined(USE_WIN32)
+#if defined(USE_IPv6)
 #define addr_len(x) ((x).sa.sa_family==AF_INET ? \
     sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6))
 #else
@@ -305,6 +333,17 @@ extern char *sys_errlist[];
 #define DEFAULT_ANY "::"
 #endif
 
+#if defined (USE_WIN32) || defined (__vms)
+#define LOG_EMERG       0
+#define LOG_ALERT       1
+#define LOG_CRIT        2
+#define LOG_ERR         3
+#define LOG_WARNING     4
+#define LOG_NOTICE      5
+#define LOG_INFO        6
+#define LOG_DEBUG       7
+#endif /* defined (USE_WIN32) || defined (__vms) */
+
 #endif /* defined COMMON_H */
 
 /* End of common.h */
